/*
 +========================================================================+
 |         (c) 2020 - PROPERTY OF ECOLE POLYTECHNIQUE - LGPL 3.0          |
 |________________________________________________________________________|
 |   '&`   |                                                              |
 |    #    |   FILE       : dirichlet.cpp                                 |
 |    #    |   VERSION    : 0.1.0                                         |
 |   _#_   |   AUTHOR(S)  : Matthieu Aussal                               |
 |  ( # )  |   CREATION   : 01.04.2020                                    |
 |  / 0 \  |   LAST MODIF : 31.10.2020                                    |
 | ( === ) |   SYNOPSIS   : Acoustical scattering using H-Matrix BEM      |
 |  `---'  |                solver for Dirichlet boundary condition       |
 +========================================================================+
 */

#include <castor/matrix.hpp>
#include <castor/smatrix.hpp>
#include <castor/hmatrix.hpp>
#include <castor/linalg.hpp>
#include <castor/graphics.hpp>
#include "fem.hpp"
#include "bem.hpp"

using namespace castor;

int main(int argc, char* argv[])
{
    // Load meshes
    matrix<double> Svtx, Pvtx;
    matrix<size_t> Stri, Ptri;
    std::tie(Stri,Svtx) = triread("/Users/aussal/Git/leprojetcastor/fembem/demo/data/","unitSphere_1e3.ply");
    std::tie(Ptri,Pvtx) = triread("/Users/aussal/Git/leprojetcastor/fembem/demo/data/","Oxyz_1e4.ply");
    
    // Graphical representation
    figure fig;
    trimesh(fig,Ptri,Pvtx);
    trimesh(fig,Stri,Svtx);
    
    // Parameters
    matrix<double> U = {-1,0,0};
    double k = 5.85;
    float tol = 1e-3;
    
    // FEM
    tic();
    femdata<double> v(Stri,Svtx,lagrangeP1,3);
    femdata<double> u(Stri,Svtx,lagrangeP1,3);
    toc();

    // Incident wave : PW(x,u) = exp(i*k*x.u)
    auto Pinc = planeWave(Pvtx,U,k);
    
    // Identity matrix
    tic();
    auto Id = mass<std::complex<double>>(v);
    toc();
    
    // Graphical representation
    figure fig2;
    spy(fig2,Id,"b","Id");
    
    // Left hand side : \int_sigma_x \int_sigma_y v(x) G(x,y) u(y) dx dy
    tic();
    auto Sbndfct = [&v,&u,&k](matrix<std::size_t> Ix, matrix<std::size_t> Iy)
    {
        return singleLayer<std::complex<double>>(v,u,k,Ix,Iy);
    };
    hmatrix<std::complex<double>> Sbnd(v.dof(),u.dof(),tol,Sbndfct);
    toc();
    disp(Sbnd);
    
    // Graphical representation
    smatrix<logical> Sf, Sc;
    tic();
    std::tie(Sf,Sc) = spydata(Sbnd);
    toc();
    figure fig3;
    spy(fig3,Sf,"r","full");
    spy(fig3,Sc,"b","compr");

    // Right hand side : - \int_sigma_x v(x) PW(x,u) dx
    auto B = - rightHandSide<std::complex<double>>(v,PWsource,U,k);

    // Solve linear system : S lambda = -PO
    hmatrix<std::complex<double>> Lh,Uh;
    tic();
    std::tie(Lh,Uh) = lu(Sbnd);
    toc();
    disp(Lh);
    disp(Uh);
    auto lambda = gmres(Sbnd,B,tol,100,Lh,Uh);
    
    // Boundary radiation
    auto Pbnd = gmres(Id,mtimes(Sbnd,lambda),tol,100) + planeWave(v.dof(),U,k);
    
    // Domain radiation
    tic();
    auto Sdomfct = [&Pvtx,&u,&k](matrix<std::size_t> Ix, matrix<std::size_t> Iy)
    {
        return singleLayer<std::complex<double>>(Pvtx,u,k,Ix,Iy);
    };
    matrix<std::complex<double>> Pdom = Pinc;
    hmatrix<std::complex<double>> Sdom(Pvtx,u.dof(),tol,Sdomfct,lambda,Pdom);
    toc();
    
    // Graphical representation
    figure fig4;
    trimesh(fig4,Ptri,Pvtx,abs(Pdom));
    trimesh(fig4,Stri,Svtx,abs(Pbnd));
    
    // Infinite radiation
    tic();
    matrix<double> theta = transpose(linspace(0,2*M_PI,1e3));
    matrix<double> Nu = horzcat(horzcat(cos(theta),sin(theta)),zeros(size(theta)));
    auto Sinffct = [&Nu,&u,&k](matrix<std::size_t> Ix, matrix<std::size_t> Iy)
    {
        return singleLayerInfinite<std::complex<double>>(Nu,u,k,Ix,Iy);
    };
    matrix<std::complex<double>> Pinf = zeros(size(Nu,1),size(lambda,2));
    hmatrix<std::complex<double>> Sinf(Nu,u.dof(),0,Sinffct,lambda,Pinf);
    toc();

    // Graphical representation
    figure fig5;
    plot(fig5,theta,transpose(log(abs(Pinf))),{"-b"});

    // Plot
    drawnow(fig);
    disp("done !");
    return 0;
}
